# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version("stag-ssg")



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

"""
Functions passed to jinja templates
"""

import logging
from dateutil.parser import parse as parse_dt
from operator import attrgetter

from stag.utils import date_sorting_key as _date_sorting

from slugify import slugify

log = logging.getLogger(__name__)


def _is_date(val):
    return isinstance(val, date)


def strftime(val, format):
    if not val:
        return ""

    try:
        if isinstance(val, (int, float, str)):
            dt = parse_dt(val)
        else:
            dt = val
        return dt.strftime(format)
    except ValueError:
        return val


def isoformat(val):
    if isinstance(val, str):
        dt = parse_dt(val)
    else:
        dt = val
    return dt.isoformat()


def rfc822format(config):
    def flt(val):
        if isinstance(val, str):
            dt = parse_dt(val)
        else:
            dt = val
        return dt.strftime(f"%a, %d %b %Y %H:%M:%S {config.timezone}")

    return flt


def sorted_pages(pages, key=None, reverse=False):
    if key in {"date", "lastmod"}:
        return sorted(pages, key=_date_sorting(key), reverse=reverse)
    return sorted(pages, reverse=reverse)


def pagetype(pages, *filetypes):
    for page in pages:
        if page.output and any(page.output.type == ft for ft in filetypes):
            yield page


def mandatory(lhs, msg):
    if not lhs:
        raise ValueError(msg)
    return lhs


def sortby(iterable, *attrs, reverse=False, default=None):
    """Sort elements by a single attribute. Attribute can be nested, in which
    case they must be given with a dot, for example: 'parent.child'. If
    attribute is missing in sorted objects, the default value is used."""
    getters = [attrgetter(attr) for attr in attrs]
    if not isinstance(default, (tuple, list)):
        default = tuple(default for _ in range(len(getters)))

    if len(getters) != len(default):
        raise ValueError("sortby: defaults must match attributes")

    def _getattr(lhs, i):
        try:
            return getters[i](lhs)
        except AttributeError:
            return default[i]

    def _key(lhs):
        return tuple(_getattr(lhs, i) for i in range(len(attrs)))

    return sorted(iterable, key=_key, reverse=reverse)


def update_env(env, config):
    globals = {"sorted_pages": sorted_pages, "slugify": slugify}
    filters = {
        "slugify": slugify,
        "strftime": strftime,
        "isoformat": isoformat,
        "rfc822format": rfc822format(config),
        "pagetype": pagetype,
        "mandatory": mandatory,
        "sortby": sortby,
    }

    env.globals.update(globals)
    env.filters.update(filters)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
import sys
import argparse
import shutil
import logging
import pkgutil
import signal
from copy import deepcopy

from stag import __version__ as _version
from stag.config import read_config
from stag.ecs import Path, Page
from stag.site import Site
from stag.writing import render
from stag.server import serve_until_changed

log = logging.getLogger(__name__)


def load_plugins_from(paths, disabled, *register_args):
    for finder, name, ispkg in pkgutil.iter_modules(paths):
        if name.startswith("_"):
            continue

        if name in disabled:
            continue

        found = finder.find_module(name)
        if found:
            mod = found.load_module()
            try:
                mod.register_plugin(*register_args)
                log.debug(f"Plugin loaded: {name}")
            except AttributeError as e:
                log.error(f"{e} - plugin will be disabled")


def load_plugins(site):
    search_paths = [
        os.path.join(os.path.dirname(__file__), "plugins"),
        site.config.plugins_path,
    ]

    disabled = site.config.plugins_disabled

    load_plugins_from(search_paths, disabled, site)
    site.signals.plugins_loaded.emit()


def build(site):
    config = site.config

    if not os.path.isdir(config.content):
        raise IOError(
            f'Not a directory: "{config.content}" (Are you inside a correct directory? Is config.content correct?).'
        )

    log.info(f"Building site to {config.output}")

    roots = [
        config.content,
        os.path.join(config.template.name, "static"),
        "static",
    ]

    site.signals.readers_init.emit(site)

    for root in roots:
        gather_files(root, site)

    site.signals.readers_finished.emit(site)

    site.signals.processors_init.emit(site)
    site.signals.processors_finished.emit(site)

    site.signals.rendering_init.emit(site)
    shutil.rmtree(config.output, ignore_errors=True)
    os.makedirs(config.output)
    render(site)
    site.signals.rendering_finished.emit(site)

    site.signals.site_finished.emit(site)


def build_cmd(args):
    config = read_config(args.config)
    override_config_with_commandline_args(config, args)

    site = Site(config=config)
    load_plugins(site)
    build(site)


def serve_cmd(args):
    config = read_config(args.config)
    override_config_with_commandline_args(config, args)

    def _exit(sig, frame):
        sys.exit(sig)

    signal.signal(signal.SIGINT, _exit)
    signal.signal(signal.SIGTERM, _exit)

    changes = 0
    while True:
        site = Site(config=config)
        load_plugins(site)

        try:
            build(site)
        except Exception as e:
            log.error(e)
            log.error("Building of site failed. You may still have the old version!")
            if args.verbosity == logging.DEBUG:
                raise
            if changes == 0:
                return

        templ = (os.path.join(config.template.name, "static"),)
        patterns = (
            "config.toml",
            f"{config.content}/**/*",
            f"{config.content}/**/.*",
            f"{templ}/*",
            f"{templ}/static/**/*",
            "static/**/*",
            config.output,
        )

        try:
            serve_until_changed(config.output, args.port, patterns)
        except Exception as e:
            log.error(e)
            log.error("Critical error detected while running HTTP server.")
            return

        log.info("")
        log.info("Change detected, regenerating.")
        changes += 1


def gather_files(srcdir, site):
    for curdir, _, files in os.walk(srcdir):
        for f in files:
            path = Path(os.path.join(curdir, f), srcdir)
            relurl = site.readers.get_path(path)
            site.make_page(relurl, source=path)


def override_config_with_commandline_args(config, args):
    for name, val in args.__dict__.items():
        if val is not None and hasattr(config, name):
            setattr(config, name, args.__dict__[name])


def prepare_args(argv):
    parser = argparse.ArgumentParser(description="Simply Stupid Static Site Generator")
    parser.set_defaults(verbosity=logging.INFO)

    parser.add_argument(
        "-c",
        "--config",
        nargs="?",
        default="config.toml",
        help="path to stag's configuration file",
    )

    parser.add_argument(
        "-C",
        "--change-directory",
        metavar="DIR",
        help="run as if stag was started in DIR instead of current working directory",
    )

    parser.add_argument(
        "-D",
        "--debug",
        action="store_const",
        const=logging.DEBUG,
        dest="verbosity",
        help="show debug messages",
    )

    parser.add_argument("--version", action="version", version=f"%(prog)s {_version}")

    sp = parser.add_subparsers(required=True, dest="subcommands")

    sp_build = sp.add_parser("build")
    sp_build.add_argument("-o", "--output", help="output directory")
    sp_build.set_defaults(func=build_cmd)

    sp_serve = sp.add_parser("serve")
    sp_serve.add_argument("-o", "--output", help="output directory")
    sp_serve.add_argument("-p", "--port", type=int, default="8000", help="HTTP port")
    sp_serve.set_defaults(func=serve_cmd)

    return parser.parse_args(argv)


def main_(argv):
    args = prepare_args(argv)
    logging.basicConfig(format="%(message)s", level=args.verbosity)

    try:
        if args.change_directory:
            os.chdir(args.change_directory)

        return args.func(args)
    except Exception as e:
        log.error(f"Critical error: {e}")
        if args.verbosity == logging.DEBUG:
            import pdb, traceback

            extype, value, tb = sys.exc_info()
            traceback.print_exc()
            pdb.post_mortem(tb)
            raise
        return 1


def main():
    return main_(sys.argv[1:])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['stag',
 'stag.plugins',
 'stag.plugins.adoc',
 'stag.plugins.macros',
 'stag.plugins.md',
 'stag.plugins.taxonomies',
 'stag.plugins.xml',
 'stag.writers']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21,<22',
 'jinja2>=3,<4',
 'markdown>=3,<4',
 'python-dateutil>=2.8.0,<3.0.0',
 'python-slugify>=5,<6',
 'tomli>=1.2,<2.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.0,<5.0']}

entry_points = \
{'console_scripts': ['stag = stag.stag:main']}

setup_kwargs = {
    'name': 'stag-ssg',
    'version': '0.8.0',
    'description': 'Deadly simple static site generator',
    'long_description': '<p align="center">\n  <a href="https://git.goral.net.pl/mgoral/stag">\n    <img alt="Logo featuring a stag" src="https://git.goral.net.pl/mgoral/stag/raw/branch/master/doc/stag.png" width="320"/>\n  </a>\n</p>\n\n# Stag\n\nStag is a simple, extensible static site generator, where almost every part\nis a plug in. It\'s almost too easy to extend it with your own\nfunctionalities.\n\n# Features\n\nOut of the box Stag comes with the following features:\n\n- pages can be generated from Markdown with enabled support for footnotes,\n  fenced code blocks and some typographic goodies.\n- support for Asciidoc (via asciidoctor)\n- generic support for file front matters\n- Jinja2 templates\n- taxonomies (e.g. tags)\n- RSS feeds\n- generation of nice URLs:\n  - _foo/index.md_ → _foo/index.html_\n  - _bar.md_ → _bar/index.html_\n- extensible with plugins and macros (shortcodes)\n\n# Installation\n\nPyPI: https://pypi.org/project/stag-ssg/\n',
    'author': 'Michal Goral',
    'author_email': 'dev@goral.net.pl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://git.goral.net.pl/mgoral/stag',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import pytest

from stag.ecs import Page, Metadata, Path
from stag.writers.template_functions import sorted_pages, sortby


class Foo:
    def __init__(self, **kw):
        for name, val in kw.items():
            setattr(self, name, val)


@pytest.fixture
def files():
    return [
        Page("/static.txt"),
        Page("/bar", metadata=Metadata(date="2021-08-09")),
        Page("/empty", metadata=Metadata()),
        Page("/foo", metadata=Metadata(date="2021-08-10")),
        Page("/baz", metadata=Metadata(date="2020-11-11")),
    ]


@pytest.fixture
def pages():
    return [
        Page("/3", metadata=Metadata(a=3)),
        Page("/1", metadata=Metadata(a=1)),
        Page("/None"),
        Page("/0", metadata=Metadata(a=0)),
        Page("/2", metadata=Metadata(a=2)),
    ]


def test_sort_default_key(files):
    srt = sorted_pages(files)
    assert srt == [files[1], files[4], files[2], files[3], files[0]]


def test_sort_default_key_reverse(files):
    srt = sorted_pages(files, reverse=True)
    assert srt == [files[0], files[3], files[2], files[4], files[1]]


def test_sort_date(files):
    srt = sorted_pages(files, key="date")
    assert srt[0] == files[4]
    assert srt[1] == files[1]
    assert srt[2] == files[3]
    # last 2 elements don't have date, so they are considered equal and their
    # order is undefined


def test_sort_date_reverse(files):
    srt = sorted_pages(files, key="date", reverse=True)
    # first 2 elements don't have date, so they are considered equal and their
    # order is undefined
    assert srt[2] == files[3]
    assert srt[3] == files[1]
    assert srt[4] == files[4]


def test_sortby(pages):
    srt = sortby(pages, "md.a", default=100)
    assert srt == [pages[i] for i in (3, 1, 4, 0, 2)]


def test_sortby_reversed(pages):
    srt = sortby(pages, "md.a", default=-1, reverse=True)
    assert srt == [pages[i] for i in (0, 4, 1, 3, 2)]


def test_sortby_many_attrs():
    pages = [
        Page("/33", metadata=Metadata(a=3, b=3)),
        Page("/3None", metadata=Metadata(a=3)),
        Page("/31", metadata=Metadata(a=3, b=1)),
    ]

    srt = sortby(pages, "md.a", "md.b", default=[0, 0])
    assert srt == [pages[i] for i in (1, 2, 0)]
